"""
A framework for managing and executing downloadable tools.
"""

import platform
import shlex
import subprocess
import sys
from abc import ABC, abstractmethod
from pathlib import Path

from . import ToolConfig, PlatformData
from .downloader import download
from .progressBar import print_progress_bar


class ExternalTool(ABC):
    def __init__(self, base_dir: Path = "./third-party", progress_bar: bool = False, lazy_setup: bool = False):
        self.base_dir = Path(base_dir)

        if not lazy_setup:
            self.setup(progress_bar)

    @property
    @abstractmethod
    def config(self) -> ToolConfig:

        raise NotImplementedError("config property must be implemented. Have you forgot to add @property decorator?")

    @property
    def tool_name(self) -> str:
        """
        Returns the name of the tool.
        """
        return self.config.tool_name

    @property
    def python(self) -> bool:
        return self.config.python

    @property
    def tool_directory(self) -> Path:
        """
        Returns the directory where the tool is installed.
        """
        return Path(self.base_dir) / Path(self.tool_name)

    def setup(self, use_progress_bar=False) -> bool:
        """
        Sets up the tool by downloading and extracting it.
        """
        if self.calculate_path().exists():
            return True

        self.tool_directory.mkdir(parents=True, exist_ok=True)
        url = self.get_platform_data().url

        # check if the tool is already downloaded
        if self.calculate_path().exists():
            return True

        if use_progress_bar:
            download(self.tool_directory, url, progress_callback=print_progress_bar)
        else:
            download(self.tool_directory, url)

        if self.python:
            requirements = (self.calculate_dir() / "requirements.txt")

            if requirements.exists():
                subprocess.check_call(
                    [sys.executable, "-m", "pip", "install", "-r",
                     requirements.resolve()])

        return True

    def get_platform_data(self) -> PlatformData:
        """
        Returns the platform data for the current operating system.

        Raises:
            ValueError: If the operating system is not supported.
        """
        system = platform.system()

        platform_data = self.config.platform_data

        for key in platform_data:
            if key.lower() == system.lower():
                system = key
                break

        if system not in platform_data:
            raise ValueError(f"Unsupported operating system: {system}")
        return platform_data[system]

    def run_command(self, cmd: str, stdout=None, stderr=None, stdin=None) -> int:
        """
        Run a command in a subprocess.

        Args:
            cmd: The command to run as a string.
            stdout: The file-like object to use as stdout.
            stdin: The file-like object to use as stdin.

        Returns:
            The exit code of the subprocess.
        """

        if not self.setup():
            raise ValueError(f"Could not set up {self.tool_name}")

        if self.python:
            cmd = f'{sys.executable} "{self.calculate_path().resolve()}" {cmd}'
        else:
            cmd = f'{self.calculate_path().resolve()} {cmd}'

        command_args = shlex.split(cmd, posix=False)

        for i, arg in enumerate(command_args):
            if arg.startswith('"') and arg.endswith('"'):
                command_args[i] = arg[1:-1]

        with subprocess.Popen(command_args, stdout=stdout, stderr=stderr, stdin=stdin, bufsize=1,
                              universal_newlines=True) as p:

            if p.stdout:
                while True:
                    line = p.stdout.readline()
                    if not line:
                        break

            exit_code = p.wait()
        return exit_code

    def calculate_path(self) -> Path:
        """
        Calculates the path of the tool based on the operating system.
        """
        directory = self.calculate_dir()
        platform_data = self.get_platform_data()
        path = directory / platform_data.executable
        return path

    def calculate_dir(self) -> Path:
        """
        Calculates the directory path of the tool based on the operating system.
        """
        subdir = self.get_platform_data().subdir
        if subdir:
            directory = self.tool_directory / subdir
        else:
            directory = self.tool_directory
        return directory



import os

import requests
import zipfile
import tempfile
from pathlib import Path


def download(tool_directory: Path, url: str, chunk_size=1024, progress_callback=None) -> None:
    """
    Downloads a zip file from the given URL and extracts it.

    :param tool_directory: The directory where the tool will be installed.
    :param url: The URL of the zip file.
    :param chunk_size: The size of the download chunks.
    :param progress_callback: A callback function to track the download progress.
    """
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    bytes_downloaded = 0

    filename = url.split('/')[-1]

    progress_context = type('', (), {})()

    with tempfile.NamedTemporaryFile(suffix=".zip", delete=False) as fp:
        temp_file_name = fp.name
        for chunk in response.iter_content(chunk_size=chunk_size):
            bytes_downloaded += len(chunk)
            fp.write(chunk)
            if progress_callback:
                progress_callback(progress_context, filename, bytes_downloaded, total_size_in_bytes)

    if progress_callback:
        progress_callback(progress_context, filename, total_size_in_bytes, total_size_in_bytes)

    with zipfile.ZipFile(temp_file_name, 'r') as zip_ref:
        zip_ref.extractall(tool_directory)

    os.remove(temp_file_name)



def print_progress_bar(context, filename, bytes_downloaded, total_size_in_bytes):
    """
    Prints the download progress.
    """

    try:
        import tqdm
    except ImportError:
        tqdm = None

    if tqdm:

        # Create a tqdm progress bar if it doesn't exist
        if not hasattr(context, "pbar"):
            context.pbar = tqdm.tqdm(desc=filename, unit='B', unit_scale=True, unit_divisor=1024, total=total_size_in_bytes, leave=False, dynamic_ncols=True)

        # Update the progress bar
        context.pbar.update(bytes_downloaded - context.pbar.n)

        # Close the progress bar if download is complete
        if bytes_downloaded >= total_size_in_bytes:
            context.pbar.close()

    else:
        print(f"Downloading {filename}: {bytes_downloaded / 1024:.2f}KB / {total_size_in_bytes / 1024:.2f}KB", end='\r')



from typing import Dict, NamedTuple

from autokit import PlatformData


class ToolConfig(NamedTuple):
    """
    A class to represent the configuration for a tool.

    Attributes
    ----------
    tool_name : str
        The name of the tool.
    python : bool
        Whether the tool should be executed by the Python interpreter or treated as an executable.
    platform_data : Dict[str, PlatformData]
        The platform data for the tool.
    """
    tool_name: str
    python: bool
    platform_data: Dict[str, PlatformData]




from pathlib import Path
from typing import Dict, NamedTuple


class PlatformData(NamedTuple):
    """
    A class to represent the data for a platform.

    Attributes
    ----------
    url : str
        The URL to download the tool from.
    subdir : Path
        The subdirectory to extract the tool to.
    executable : Path
        The path to the executable, relative to the subdir.
    """
    url: str
    subdir: Path
    executable: Path



